/* 
   sitecopy, for managing remote web sites. WebDAV client routines.
   Copyright (C) 1998-99, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   $Id: httpdav.c,v 1.29.2.1 1999/05/03 10:21:56 joe Exp $
*/

/* This file is a collection of routines to implement a basic WebDAV
 * client, including an HTTP/1.1 client. 
 * Transparantely supports basic and digest authentication.
 */

/* TODO:
 * String overflow checking.
 * Better error condition handling in http_request - SPLIT UP http_request.
 * Handle Authentication-Info: response header... but first, find a 
 * server that actually sends it.
 */

/* How it works....
 * http_request_init, then http_request(), basically.
 */

#include "config.h"

#include <sys/types.h>

#include <sys/stat.h>
#include <sys/time.h>
#ifdef __EMX__
#include <sys/select.h>
#endif
#include <netinet/in.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifndef HAVE_SNPRINTF
#include <snprintf.h>
#endif

#include "frontend.h" /* for fe_transfer_progess */
#include "protocol.h"
#include "httpdav.h"
#include "common.h"
#include "socket.h"
#include "md5.h"

/* Connection information... */
int http_sock;
/*bool http_persistent;  Whether persistent connections work */
/* For persistent connections, this is set to false when the socket
 * is closed by the server, so the next request has to open it again.
 * For non-persistent connections, this is not used */
bool http_connected;

/* Time in seconds to wait for the server to give us a 100 Continue
 * after submitting a PUT with an "Expect: 100-continue" header.
 */
#define HTTP_EXPECT_TIMEOUT 15
/* 100-continue only used if size > HTTP_EXPECT_MINSIZ */
#define HTTP_EXPECT_MINSIZE 512

/* Whether the current server respects the Expect: 100-continue header */
int http_expect_works; /* == 0 if it we don't know, 1 if it does,
			    -1 if it doesn't support 100-continue */

bool http_webdav_server = false;
bool http_init_checks = true;

#ifdef HAVE_LIBEXPAT
/* WebDAV Fetch mode */

#include <xmlparse.h>

/* Prototypes */
void dav_xml_startelm( void *userdata, const char *tag, const char **atts );
void dav_xml_endelm( void *userdata, const char *tag );
void dav_xml_cdata( void *userdata, const char *cdata, int len );

#endif /* HAVE_LIBEXPAT */

/* Handy macro to free things. */
#define DOFREE(x) if( x!=NULL ) free( x )

#define EOL "\r\n"
#define HTTP_PORT 80

const char *http_useragent = PACKAGE "/" VERSION;

/* We need to remember the remote host and port even after connecting
 * the socket e.g. so we can form the Destination: header */
char *http_remotehost;
struct in_addr http_remoteaddr;
int http_remoteport;

/* The transfer encoding types */
enum http_te_t {
    /* TODO: Support other types */
    http_te_none,
    http_te_chunked,
    http_te_unknown
};

typedef enum {
    http_auth_basic,
    http_auth_digest,
    http_auth_none
} http_auth_t;

/* HTTP Authorization. 
 *   http_auth holds the type of authorization last used.
 *   http_auth_cookie holds the last cookie used for basic auth.
 *   http_username and http_password hold the username and password
 *     last used.
 *   http_auth_realm, _opaquea, _nonce are all used by digest auth.
 */
http_auth_t http_auth;
char *http_username, *http_password;
char *http_auth_cookie;
char *http_auth_realm, *http_auth_opaque, *http_auth_nonce;

int http_mkdir_works;

/* Holds the error message to be passed up */
char http_error[BUFSIZ];

/* Private prototypes */

/* Header parser for OPTIONS requests. */
void http_options_parsehdr( const char *name, const char *value );

int http_parse_status( http_req_t *req, char *status_line );

/* Opens the connection to the remote server.
 * Returns:
 *  PROTO_OK       on success
 *  PROTO_CONNECT  if the socket couldn't be connected
 */
int http_open( void );

/* Closes the connection.
 * Always returns PROTO_OK
 */
int http_close( void );

/* This doesn't really connect to the server.
 * Returns
 *  PROTO_LOOKUP if the hostname of the server could not be resolved
 *  PROTO_LOCALHOST if the hostname of the local machine could not be
 *    found
 *  PROTO_CONNECT if the socket could not be connected
 *  PROTO_OK if the connection was made successfully.
 */
int http_init( const char *remote_root,
	       const char *hostname, const int port,
	       const char *username, const char *password ) {
    int ret;
    /* Look up the remote host name */
    if( http_remotehost!=NULL ) free( http_remotehost );
    http_remotehost = strdup( hostname );
    http_remoteport = port;
    if( host_lookup( http_remotehost, &http_remoteaddr ) )
	return PROTO_LOOKUP;
    http_connected = false;
    http_expect_works = 0; /* we don't know yet */
    http_mkdir_works = 0; /* we don't know yet */
    /* Initialize the authentication variables */
    DOFREE( http_auth_nonce );
    DOFREE( http_auth_realm );
    DOFREE( http_auth_opaque );
    http_auth = http_auth_none;
    http_username = strdup( username );
    http_password = strdup( password );
    ret = http_open();
    /* Drop out if that failed, or they don't want the OPTIONS */
    if( (!http_init_checks) || (ret != PROTO_OK) )
	return ret;
    /* Capability discovery... we don't care whether this
     * actually works or not, we just want http_webdav_server
     * set appropriately. */
    (void) http_options( remote_root );
    return PROTO_OK;
}

int http_finish( void ) {
    free( http_username );
    free( http_password );
    if( http_connected ) http_close( );
    return PROTO_OK;
}

/* This parses the response status line, and plonks the numeric
 * status value in status.
 * Returns: PROTO_OK on success, PROTO_ERROR otherwise */
int http_parse_status( http_req_t *req, char *status_line ) {
    char *part;
    DEBUG( DEBUG_HTTP, "HTTP response line: %s", status_line );
    /* Save the line for error codes later */
    memset( http_error, 0, BUFSIZ );
    strncpy( http_error, status_line, BUFSIZ );
    /* Strip off the CRLF for the status line */
    if( (part = strrchr( http_error, '\r' )) != NULL )
	*part = '\0';
    /* Parse the status code */
    part = strtok( status_line, " " ); /* "HTTP/1.1" */
    if( part==NULL ) return PROTO_ERROR;
    part = strtok( NULL, " " ); /* xxx */
    if( part==NULL ) return PROTO_ERROR;
    req->status = atoi( part );
    DEBUG( DEBUG_HTTP, "HTTP reponse code: %d\n", req->status );
    return PROTO_OK;
}

/* Add the authorization headers to the request. 
 * FIXME: Check whether this leaks like a sieve.
 */
void http_auth_request( http_req_t *req, char *headers ) {
    char tmp[BUFSIZ];
    char *a1, *md5_a1, *a2, *x, *y, *z, *temp, *auth_response,
	*v, *w, *ei; /* yarg, that enough? */

    switch( http_auth ) {
    case http_auth_basic:
	strcat( headers, "Authorization: Basic " );
	strcat( headers, http_auth_cookie );
	strcat( headers, EOL );
	break;
    case http_auth_digest:
/* RFC2069 Says:
--- begins ---
     response-digest     =
          <"> < KD ( H(A1), unquoted nonce-value ":" H(A2) > <">

     A1             = unquoted username-value ":" unquoted realm-value
                                                ":" password
     password       = < user's password >
     A2             = Method ":" digest-uri-value
--- ends ---

Which means, in English...
   A1 = username + ":" + realm + ":" password
   A2 = method + ":" + uri
   x = nonce + ":" + MD5( A2 )
   y = MD5( A1 ) + ":" + x
   z = MD5( y )
   ( and response-digest = "\"" + z + "\"", but we only need to 
   calculate z, which is put into http_auth_response );
*/
	a1 = malloc( strlen( http_username ) +
		     strlen( http_auth_realm ) +
		     strlen( http_password ) + 3 );
	strcpy( a1, http_username );
	strcat( a1, ":" );
	strcat( a1, http_auth_realm );
	strcat( a1, ":" );
	strcat( a1, http_password );
	DEBUG( DEBUG_HTTP, "A1: %s\n", a1 );
	a2 = malloc( strlen( req->method ) +
		     strlen( req->uri ) + 2 );
	strcpy( a2, req->method );
	strcat( a2, ":" );
	strcat( a2, req->uri );
	DEBUG( DEBUG_HTTP, "A2: %s\n", a2 );
	temp = md5_hex( a2, strlen( a2 ) );
	free( a2 );
	x = malloc( strlen( http_auth_nonce ) +
		    strlen( temp ) + 2 );
	strcpy( x, http_auth_nonce );
	strcat( x, ":" );
	strcat( x, temp );
	free( temp );
	DEBUG( DEBUG_HTTP, "x: %s\n", x );
	md5_a1 = md5_hex( a1, strlen( a1 ) );
	free( a1 );
	y = malloc( strlen( md5_a1 ) +
		    strlen( x ) + 2 );
	strcpy( y, md5_a1 ); 
	strcat( y, ":" );
	strcat( y, x );
	/* take a DEEP breath */
	auth_response = md5_hex( y, strlen( y ) );
	DEBUG( DEBUG_HTTP, "Response digest: %s\n", temp );
	free( x );
	free( y );
	snprintf( tmp, BUFSIZ, "Authorization: Digest "
		  "username=\"%s\", realm=\"%s\", "
		  "nonce=\"%s\", uri=\"%s\", response=\"%s\"",
		  http_username,
		  http_auth_realm,
		  http_auth_nonce,
		  req->uri,
		  auth_response );

	free( auth_response );

/* Now for the entity digest. RFC2069 says:
--- begins ---
entity-digest = <"> KD (H(A1), unquoted nonce-value ":" Method ":"
                           date ":" entity-info ":" H(entity-body)) <">
       ; format is <"> *LHEX <">

date = = rfc1123-date            ; see section 3.3.1 of [2]
entity-info = H(
          digest-uri-value ":"
          media-type ":"         ; Content-type, see section 3.7 of [2]
          *DIGIT ":"             ; Content length, see 10.12 of [2]
          content-coding ":"     ; Content-encoding, see 3.5 of [2]
          last-modified ":"      ; last modified date, see 10.25 of [2]
          expires                ; expiration date; see 10.19 of [2]
          )

last-modified   = rfc1123-date  ; see section 3.3.1 of [2]
expires         = rfc1123-date
--- ends ---

Since we do not pass last-modified, content-encoding, expires, or
content-encoding headers, we can skip those fields.

So what we do is this:

   ei = "::" + body_size + ":::"
   y = MD5( body )
   date = date from date header (in format specified by RFC1123)
   z = nonce + ":" + method + ":" + date + ":" + ei + ":" + y
   md5_a1 = MD5( A1 )    (from above)
   v = md5_a1 + ":" + z
   w = MD5( v )   where w is the unquoted entity-digest
*/
	if( req->body!=http_body_none ) {
	    temp = malloc( BUFSIZ );
	    snprintf( temp, BUFSIZ, "%d", req->body_size );
	    ei = malloc( strlen(temp) + 5 + 1 );
	    strcpy( ei, "::" );
	    strcat( ei, temp ); free( temp );
	    strcat( ei, ":::" );
	    DEBUG( DEBUG_HTTP, "Entity-info: %s\n", ei );
	    if( req->body == http_body_file ) {
		y = md5_hex_stream( req->body_file );
		rewind( req->body_file ); /* leave it at the beginning */
	    } else {
		y = md5_hex( req->body_buffer, req->body_size );
	    }
	    z = malloc( strlen( http_auth_nonce ) + strlen( req->method ) +
			strlen( req->date ) + strlen( ei ) + strlen( y ) + 5 );
	    strcpy( z, http_auth_nonce );
	    strcat( z, ":" );
	    strcat( z, req->method );
	    strcat( z, ":" );
	    strcat( z, req->date );
	    strcat( z, ":" );
	    strcat( z, ei ); free( ei );
	    strcat( z, ":" );
	    strcat( z, y ); free( y );
	    DEBUG( DEBUG_HTTP, "Z: %s\n", z );
	    v = malloc( strlen( md5_a1 ) + strlen( z ) + 2 );
	    strcpy( v, md5_a1 );
	    strcat( v, ":" );
	    strcat( v, z ); free( z );
	    DEBUG( DEBUG_HTTP, "V: %s\n", v );
	    w = md5_hex( v, strlen( v ) );
	    free( v );
	    DEBUG( DEBUG_HTTP, "W: %s\n", w );
	    /* Now lob it on the end */
	    strcat( tmp, ", digest=\"" );
	    strcat( tmp, w ); free( w );
	    strcat( tmp, "\"" );
	}
	free( md5_a1 );

	if( http_auth_opaque!=NULL ) {
	    strcat( tmp, ", opaque=\"" );
	    strcat( tmp, http_auth_opaque );
	    strcat( tmp, "\"" );
	}
	strcat( tmp, EOL );
	strcat( headers, tmp );
	break;
    default:
	break;
    }
}

/* This is called when a WWW-Authenticate header is received in the response.
 * It is used to perform any necessary computation required for presenting
 * authorisation information on the next (retry-with-auth) request.
 * i.e., to work out http_auth_cookie.
 * Basic auth is simple - we just take the string
 *   username:password
 * and Base64 it.
 * Digest auth is complex - read RFC2069.
 */
void http_auth_challenge( char *challenge ) {
    /* Parse the authentication challenge */
    char auth_type[BUFSIZ], auth_parms[BUFSIZ];
    char *part, *temp, *key, *val, *keypnt;
    size_t len, len2;
    /* FIXME: Decent parser.
     * Here, we split the entire challenge into the challenge part and the
     * other parameters - everything before, and after the first space char,
     * respectively. 
     */
    len = strcspn( challenge, " " );
    DEBUG( DEBUG_HTTP, "Challenge: [%s]\n", challenge );
    if( len == 0 ) {
	DEBUG( DEBUG_HTTP, "Invalid challenge header: [%s]\n", challenge );
	return;
    }
    strncpy( auth_type, challenge, len );
    *(auth_type+len) = '\0'; /* null-term */
    strcpy( auth_parms, challenge+len+1 );
    /* Now process it */
    DEBUG( DEBUG_HTTP, "Challenge: type [%s] parms [%s]\n", auth_type,
	   auth_parms );
    if( strcasecmp( auth_type, "basic" ) == 0 ) {
	if( http_auth==http_auth_digest ) {
	    /* We're being spoofed, don't hand out the password.
	     * FIXME: Get this to the user, somehow. */
	    DEBUG( 0, 
"WARNING: Spoof attack probable. The server has switched to using basic\n"
"authentication from digest authenticaion, which may be an attempt to\n"
"discover your password. Basic auth will NOT be used.\n" );
	    return;
	}
	DEBUG( DEBUG_HTTP, "Using basic authentication.\n" );
	http_auth = http_auth_basic;
	temp = malloc( strlen(http_username) + strlen(http_password) + 2 );
	strcpy( temp, http_username );
	strcat( temp, ":" );
	strcat( temp, http_password );
	DEBUG( DEBUG_HTTP, "Uname/password pair: [%s]\n", temp );
	http_auth_cookie = base64( temp );
	DEBUG( DEBUG_HTTP, "Basic auth cookie: %s\n", http_auth_cookie );
	http_auth = http_auth_basic;
	free( temp );
    } else if( strcasecmp( auth_type, "digest" ) == 0 ) {
	/* Digest authentication */
	DEBUG( DEBUG_HTTP, "Trying digest authentication.\n" );
	/* FIXME: Another decent parser needed */
	part = strtok( auth_parms, "," );
	/* Chuck out the old nonce + realm */
	DOFREE( http_auth_nonce );
	DOFREE( http_auth_realm );
	DOFREE( http_auth_opaque );
	while( part!=NULL ) {
	    /* Get the first bit before the = */
	    len = strcspn( part, "=" );
	    key = malloc( len + 1 );
	    keypnt = key;
	    strncpy( key, part, len );
	    *(key+len) = '\0'; /* null-terminate */
	    while( *keypnt==' ' ) keypnt++; /* skip leading whitespace */
	    /* Value is ALWAYS a quoted string */
	    len2 = strlen( part ) - (len+3);
	    if( len2 > 0 ) {
		val = malloc( len2 + 1 );
		strncpy( val, part+len+2, len2 );
		*(val+len2) = '\0'; /* null-terminate */
		DEBUG( DEBUG_HTTP, "Got: [%s] - [%s]\n", keypnt, val );
		if( strcasecmp( keypnt, "nonce" ) == 0 ) {
		    http_auth_nonce = val;
		} else if( strcasecmp( keypnt, "realm" ) == 0 ) {
		    http_auth_realm = val;
		} else if( strcasecmp( keypnt, "opaque" ) == 0 ) {
		    http_auth_opaque = val;
		} else {
		    /* it's no use, chuck it */
		    free( val );
		}
	    }
	    part = strtok( NULL, "," );
	    free( key );
	}
	DEBUG( DEBUG_HTTP, "Nonce: %s, realm: %s\n",
	       http_auth_nonce, http_auth_realm );    
	if( (http_auth_nonce != NULL) && (http_auth_realm != NULL) ) {
	    http_auth = http_auth_digest;
	} else {
	    http_auth = http_auth_none;
	    DEBUG( DEBUG_HTTP, "Could not use digest authentication.\n" );
	}
    } else {
	http_auth = http_auth_none;
	DEBUG( DEBUG_HTTP, "Unrecognised authentication method.\n" );
    }
}

/* Sends the body down the socket.
 * Returns PROTO_OK on success, PROTO_ERROR otherwise */
int http_req_sendbody( http_req_t *req ) {
    int ret;
    switch( req->body ) {
    case http_body_file:
	ret = transfer( fileno(req->body_file), http_sock, req->body_size );
	rewind( req->body_file ); /* since we may have to send it again */
	break;
    case http_body_buffer:
	DEBUG( DEBUG_HTTP, "Sending body:\n%s\n", req->body_buffer );
	ret = send_string( http_sock, req->body_buffer );
	break;
    default:
	DEBUG( DEBUG_HTTP, "Argh in http_req_sendbody!" );
	ret = -1;
    }
    if( ret == -1 ) { 
	/* transfer failed */
	return PROTO_ERROR;
    } else {
	return PROTO_OK;
    }
}

/* Deal with the body size */
int http_req_bodysize( http_req_t *req ) {
    struct stat bodyst;
    /* Do extra stuff if we have a body */
    switch( req->body ) {
    case http_body_file:
	/* Get file length */
	if( fstat( fileno(req->body_file), &bodyst ) < 0 ) {
	    /* Stat failed */
	    DEBUG( DEBUG_HTTP, "Stat failed: %s\n", strerror( errno ) );
	    return PROTO_ERROR;
	}
	req->body_size = bodyst.st_size;
	break;
    case http_body_buffer:
	req->body_size = strlen( req->body_buffer );
	break;
    default:
	/* No body, so no size. */
	return PROTO_OK;
    }
    if( req->body != http_body_none ) {
	char tmp[BUFSIZ];
	/* Add the body length header */
	snprintf( tmp, BUFSIZ, "Content-Length: %d" EOL, req->body_size );
	strcat( req->headers, tmp );
    }
    return PROTO_OK;
}

void http_strcat_remoteserver( char *str ) {
    strcat( str, http_remotehost );
    /* Only add the port if it isn't 80 */
    if( http_remoteport != HTTP_PORT ) {
	static char buffer[128];
	snprintf( buffer, 128, ":%d", http_remoteport );
	strcat( str, buffer );
    }
}

/* Lob the User-Agent, connection and host headers in to the request
 * headers */
void http_req_fixedheaders( http_req_t *req ) {
    strcat( req->headers, "User-Agent: " );
    strcat( req->headers, http_useragent );
    strcat( req->headers, EOL 
	    "Connection: Keep-Alive" EOL 
	    "Host: " );
    http_strcat_remoteserver( req->headers );
    strcat( req->headers, EOL );
}

/* Initializes the request with given method and uri */
void http_request_init( http_req_t *req, 
			const char *method, const char *uri ) {
    /* Clear it out */
    memset( req, 0, sizeof( http_req_t ) );

    /* Add in the fixed headers */
    http_req_fixedheaders( req );

    /* Set the standard stuff */
    req->method = method;
    req->uri = uri;
    
    req->body_callback = NULL;
    req->body = http_body_none;
    
}


/* The HTTP/1.x request/response mechanism 
 *
 * Returns:
 *   PROTO_OK if the request was made (not related to status code)
 *   PROTO_ERROR if the request could not be made
 * The STATUS CODE is placed in req->status. The error string is
 * placed in http_error.
 * 
 * TODO: This should be chopped up into smaller chunks, and get rid of
 * the horrid goto's.  
 */
int http_request( http_req_t *req ) {
    char tmp[BUFSIZ], buffer[BUFSIZ], *part, *name, *value;
    int ret, length, attempt, con_attempt;
    fd_set sockfd;
    struct timeval tv;
    bool using_expect, /* whether we have sent a Expect: 100 header */
	close_connection,
	dead_connection;
    enum http_te_t encoding;

#ifdef USE_BROKEN_PROPFIND
    bool is_propfind = (strcmp( req->method, "PROPFIND" ) == 0);
#endif

#define HTTP_FATAL_ERROR(a...) {				\
	DEBUG( DEBUG_HTTP, a );					\
	ret = PROTO_ERROR;					\
	close_connection = true;				\
	goto http_request_finish_proc;				\
    }

    /* Initialization... */
    DEBUG( DEBUG_HTTP, "Request started...\n" );
    strcpy( http_error, "Unknown error." );
    ret = PROTO_OK;

    if( http_req_bodysize( req ) != PROTO_OK )
	return PROTO_ERROR;

    /* I shall try this only twice...
     * First time, with default authentication stuff (either, what we
     * did last time, or none at all), then with up-to-the-minute
     * what-the-server-requested authentication stuff. */

    attempt = con_attempt = 1;

    do {
	char request[REQSIZ]; 

	/* The Request-Line */
	snprintf( request, REQSIZ, 
		  "%s %s HTTP/1.1" EOL, req->method, req->uri );
	/* The caller-supplied headers */
	strcat( request, req->headers );

	/* For the Date: header, and digest auth */
	req->date = rfc1123_date( time(NULL) );

	if( http_auth!=http_auth_none ) {
	    /* Add the authorization headers in */
	    http_auth_request( req, request );
	}

	/* Now handle the body. */
	using_expect = false;
	if( req->body!=http_body_none ) {
	    /* Add the date header */
	    strcat( request, "Date: " );
	    strcat( request, req->date );
	    strcat( request, EOL );
	    if( (http_expect_works > -1) &&
		(req->body_size > HTTP_EXPECT_MINSIZE) 
#ifdef USE_BROKEN_PROPFIND
		/* ... definitely NOT if we're doing PROPFIND */
		&& (!is_propfind)
#endif /* USE_BROKEN_PROPFIND */
		) {
		/* Add Expect: 100-continue. */
		strcat( request, "Expect: 100-continue" EOL );
		using_expect = true;
	    }
	}

	free( req->date ); /* that's enough of that */

	/* Final CRLF */
	strcat( request, EOL );
	
	/* Now send the request */

	/* Open the connection if necessary */
	if( !http_connected ) {
	    if( (ret = http_open()) != PROTO_OK )
		return ret;
	}

	dead_connection = false;

#ifdef USE_BROKEN_PROPFIND
	if( !is_propfind ) {
#endif
	/* Send the headers */
	DEBUG( DEBUG_HTTP, "Sending request headers:\n%s", request );
	if( send_string( http_sock, request ) < 0 ) {
	    dead_connection = true;
	    HTTP_FATAL_ERROR( "Could not send request!\n" );
	}

	DEBUG( DEBUG_HTTP, "Request sent.\n" );
	
#ifdef USE_BROKEN_PROPFIND
	}
#endif /* USE_BROKEN_PROPFIND */

	/* Now, if we are doing a Expect: 100, hang around for a short
	 * amount of time, to see if the server actually cares about the 
	 * Expect and sends us a 100 Continue response if the request
	 * is valid, else an error code if it's not. This saves sending
	 * big files to the server when they will be rejected.
	 */
	
	if( using_expect ) {
	    DEBUG( DEBUG_HTTP, "Waiting for response...\n" );
	    /* And wait for something to happen for a while */
	    /* Set up the fd set for the select() */
	    FD_ZERO( &sockfd );
	    FD_SET( http_sock, &sockfd );
	    /* Set the timeout appropriately */
	    tv.tv_sec = HTTP_EXPECT_TIMEOUT;
	    tv.tv_usec = 0;
	    ret = select( http_sock+1, &sockfd, NULL, NULL, &tv );
	    
	    switch( ret ) {
	    case -1: /* error */
		HTTP_FATAL_ERROR( "Wait (select) went wrong: %s\n",
				  strerror( errno ) );
		break;
	    case 1: /* we got us a response! */
		DEBUG( DEBUG_HTTP, "Wait got data.\n" );
		http_expect_works = 1; /* it works - use it again */
		break;
	    case 0: 
		/* Timed out - i.e. Expect: ignored. There is a danger
		 * here that the server DOES respect the Expect: header,
		 * but was going SO slowly that it didn't get time to
		 * respond within HTTP_EXPECT_TIMEOUT.
		 * TODO: while sending the body, check to see if the
		 * server has sent anything back - if it HAS, then
		 * stop sending - this is a spec compliance SHOULD */
		DEBUG( DEBUG_HTTP, "Wait timed out.\n" );
		http_expect_works = -1; /* don't try that again */
		/* and give them the body */
		if( http_req_sendbody( req ) != PROTO_OK )
		    HTTP_FATAL_ERROR( "Could not send body.\n" );
		break;
	    }
	} else if( req->body != http_body_none ) {
#ifdef USE_BROKEN_PROPFIND
	    if( is_propfind ) {
		char *pfbuf;
		DEBUG( DEBUG_HTTP, "Broken PROPFIND handler...\n" );
		pfbuf = malloc( strlen(request) + req->body_size + 1 );
		memset( pfbuf, 0, strlen(request) + req->body_size + 1 );
		strcpy( pfbuf, request );
		strcat( pfbuf, req->body_buffer );
		DEBUG( DEBUG_HTTP, "Sending PROPFIND request...\n" );
		if( send_string( http_sock, pfbuf ) < 0 ) {
		    HTTP_FATAL_ERROR( "Could not send request.\n" );
		}
		DEBUG( DEBUG_HTTP, "Sending newline.\n" );
		if( send_line( http_sock, "" ) < 0 ) {
		    HTTP_FATAL_ERROR( "Could not send newline.\n" );
		}
	    } else {
		
#endif /* USE_BROKEN_PROPFIND */
	    /* Just chuck the file down the socket */
	    DEBUG( DEBUG_HTTP, "Sending body...\n" );
	    if( http_req_sendbody( req ) == PROTO_ERROR )
		HTTP_FATAL_ERROR( "Could not send body.\n" );
	    /* FIXME: This. Should it be here? */
	    DEBUG( DEBUG_HTTP, "Sending newline.\n" );
	    if( send_line( http_sock, "" ) < 0 ) {
		HTTP_FATAL_ERROR( "Could not send newline.\n" );
	    }
	    DEBUG( DEBUG_HTTP, "Body sent.\n" );
#ifdef USE_BROKEN_PROPFIND
	    }
#endif /* USE_BROKEN_PROPFIND */
	}
	
	/* Now, we have either:
	 *   - Sent the header and body, or
	 *   - Sent the header incl. Expect: line, and got some response.
	 * In any case, we get the status line of the response.
	 */
	
	/* HTTP/1.1 says we MUST be able to accept any number of
	 * 100 (Continue) responses prior to the normal response.
	 * So loop while we get them.
	 */
	
	do {
	    if( read_line( http_sock, tmp, BUFSIZ ) < 0 ) {
		dead_connection = true;
		HTTP_FATAL_ERROR( "Could not read status line.\n" );
	    }
	    
	    DEBUG( DEBUG_HTTP, "[Status Line] < %s", tmp );
	    
	    /* Got the status line - parse it */
	    if( http_parse_status( req, tmp ) == PROTO_ERROR )
		HTTP_FATAL_ERROR( "Could not parse status line.\n" );

	    if( req->status == 100 ) {
		DEBUG( DEBUG_HTTP, "Got 100-continue.\n" )
		/* Skip any headers, we don't need them */
		do {
		    if( read_line( http_sock, tmp, BUFSIZ ) < 0 )
			HTTP_FATAL_ERROR( "Could not read header.\n" );
		    DEBUG( DEBUG_HTTP, "[Ignored header] < %s", tmp );
		} while( strcmp( tmp, EOL ) != 0 );
	
		if( using_expect ) {
		    /* We are using Expect: 100, and we got a 100-continue 
		     * return code... send the request body */
		    DEBUG( DEBUG_HTTP, "Got continue... sending body now.\n" );
		    if( http_req_sendbody( req ) != PROTO_OK )
			HTTP_FATAL_ERROR( "Could not send body.\n" );
		    DEBUG( DEBUG_HTTP, "Body sent.\n" );
		}
	    }
	} while( req->status == 100 );
	
	/* We've got the real status line... now get the headers */
	
	length = -1;
	encoding = http_te_unknown;
	close_connection = false;
	
	/* Now read the rest of the header... up to the next blank line */
	while( read_line( http_sock, buffer, BUFSIZ ) > 0 ) {

	    DEBUG( DEBUG_HTTP, "[Header:%d] < %s", strlen(buffer), buffer );
	    if( strcmp( buffer, EOL ) == 0 ) {
		DEBUG( DEBUG_HTTP, "CRLF: End of headers.\n" );
		break;
	    }
	    /* Now parse the header line. This is all a bit noddy. */
	    name = strtok( buffer, ":" );
	    if( name != NULL ) {
		value = strtok( NULL, ":" );
		if( value != NULL ) {
		    /* Strip leading whitespace, and trailing CRLF */
		    while( *value!='\0' && *value==' ' ) value++;
		    if( (part = strchr( value, '\r' )) != NULL ) 
			*part = '\0';
		    if( (part = strchr( value, '\n' )) != NULL ) 
			*part = '\0';
		} else {
		    /* No value given */
		    value = "";
		}

		if( strcasecmp( name, "Content-Length" ) == 0 ) {
		    length = atoi( value );
		} else if( strcasecmp( name, "Transfer-Encoding" ) == 0 ) {
		    if( strcasecmp( value, "chunked" ) == 0 ) {
			encoding = http_te_chunked;
		    } else {
			encoding = http_te_unknown;
		    }
		} else if( strcasecmp( name, "Connection" ) == 0 ) {
		    if( strcasecmp( value, "close" ) == 0 ) {
			close_connection = true;
		    }
		} else if( strcasecmp( name, "WWW-Authenticate" ) == 0 ) {
		    /* Parse the authentication challenge */
		    http_auth_challenge( value );
		} else if( req->hdrs_callback != NULL ) {
		    (*req->hdrs_callback)( name, value );
		}
	    }
	}

	/* Body length calculation, bit icky.
	 * Here, we set:
	 * length==-1 if we DO NOT know the exact body length
	 * length>=0 if we DO know the body length.
	 *
	 * RFC2068, section 4.3: 
	 * NO body is returned if the method is HEAD, or the resp status
	 * is 204 or 304
	 */
	if( (strcmp( req->method, "HEAD" ) == 0 ) ||
	    req->status==204 ||
	    req->status==304 ) {
	    length = 0;
	} else {
	    /* RFC2068, section 4.4: if we have a transfer encoding
	     * and a content-length, then ignore the content-length. */
	    if( ( length>-1 ) && (encoding!=http_te_unknown ) ) {
		length = -1;
	    }
	}

	/* Now, read the body */

	 if( length == 0 ) {
	    /* Doing a HEAD, [23]04, or got Content-Length: 0.
	     * Nowt to do. */
	    DEBUG( DEBUG_HTTP, "No response body to fetch.\n" );
	} else if( encoding==http_te_chunked ) {
	    /* The response body is encoded using 'chunks'.
	     * It goes:  `SIZE CRLF CHUNK CRLF SIZE CRLF CHUNK CRLF ...'
	     * ended by a `SIZE CHUNK 0 CRLF', a 0-byte chunk */
	    int chunk_size; /* the size of the next chunk */
	    char *chunk; /* the data of the next chunk */

	    DEBUG( DEBUG_HTTP, "Reading chunked body.\n" );

	    if( req->resp_body_size )
		*req->resp_body_size = 0;

	    do {
		/* Pity the HTTP spec doesn't specify a maximum
		 * chunk size */
		if( read_line( http_sock, buffer, BUFSIZ ) < 0 )
		    HTTP_FATAL_ERROR( "Could not read chunk size.\n" );
		DEBUG( DEBUG_HTTP, "[Chunk Size] < %s", buffer );
		sscanf( buffer, "%x", &chunk_size );
		DEBUG( DEBUG_HTTP, "Got chunk size: %d\n", chunk_size );
		if( req->resp_body_size ) /* best we can do */
		    *req->resp_body_size += chunk_size;
		if( chunk_size > 0 ) {
		    chunk = malloc( chunk_size + 1 );
		    if( read_data( http_sock, chunk, chunk_size ) < 0 ) {
			/* It broke */
			free( chunk );
			HTTP_FATAL_ERROR( "Could not read chunk data.\n" );
		    }
		    *(chunk+chunk_size) = '\0';
		    DEBUG( DEBUG_SOCKET, "<<< Chunk:\n%s\n", chunk );
		    /* Do they want to have a look? */
		    if( req->body_callback )
			(*req->body_callback)( chunk, chunk_size );
		    /* Now chuck the chunk, we don't want it */
		    free( chunk ); 
		    /* Skip CRLF */
		    if( read_line( http_sock, buffer, BUFSIZ ) < 0 ) {
			/* It broke */
			HTTP_FATAL_ERROR( "Could not read CRLF.\n" );
		    }
		} else {
		    /* Finished getting body */
		    if( req->body_callback )
			(*req->body_callback)( NULL, 0 );
		}
	    } while( chunk_size > 0 ); /* loop until 0-length chunk */
	    /* Read in that last line */
	    DEBUG( DEBUG_HTTP, "Reading trailing newline.\n" );
	    if( read_line( http_sock, buffer, BUFSIZ ) < 0 ) {
		/* It broke, kill it */
		HTTP_FATAL_ERROR( "Could not reading trailing CRLF.\n" );
	    }
	    DEBUG( DEBUG_HTTP, "[Trail] < %s", buffer );

	} else if( length > 0 ) {
	    char block[BUFSIZ+1]; /* +1 so we can null-term it */
	    int block_size;
	    /* No encoding, but we've got a Content-Length. */
	    if( req->resp_body_size )
		*req->resp_body_size = length;
	    while( length > 0 ) {
		DEBUG( DEBUG_HTTP,
		      "Reading block of size: %d\n", min(BUFSIZ,length) );
		block_size = read( http_sock, block, min(BUFSIZ,length) );
		if( block_size < 1 )
		    HTTP_FATAL_ERROR( "Block read failed: %s.\n", 
				      strerror(errno) );
		block[block_size] = '\0';
		/* Let them inspect the block */
		DEBUG( DEBUG_HTTP, "[Body block: Size %d] <<<\n%s\n", 
		       block_size, block );
		length -= block_size;
		DEBUG( DEBUG_HTTP, "Body length remaining: %d\n", length );
		if( req->body_callback )
		    (*req->body_callback)( block, block_size );
	    }
	    if( req->body_callback )
		(*req->body_callback)( block, 0 );
	    DEBUG( DEBUG_HTTP, "Read body successfully.\n" );
	} else {
	    /* TODO: Handle this by read-to-socket-close. */
	    DEBUG( DEBUG_HTTP, "Response body is lost on me.\n" );
	    /* And the connection will have to be closed */
	    close_connection = true;
	}
	
	/* If they got this far, it's worked */
	ret = PROTO_OK;
    
http_request_finish_proc:
	/* Now, do we close the connection? */
	if( close_connection ) {
	    DEBUG( DEBUG_HTTP, "Forced connection close.\n" );
	    http_close( );
	}
    
	if( !dead_connection && (http_auth == http_auth_none) )
	    /* We can't authenticate... give up.
	     * http_auth will have been set by http_auth_challenge
	     * by know if it found out a way in which we can 
	     * authenticate with the server. */
	    break;
	
	/* Now, do that all *again* if it didn't work.
	 * Otherwise, give up */

    } while( (dead_connection && (++con_attempt<4)) || ((++attempt<3) && (req->status==401)) );
    
    DEBUG( DEBUG_HTTP, "Request ends. Last status: %s\n", http_error );

    if( req->status == 0 ) {
	/* Baaddd, since they're going to /100 it */
	DEBUG( DEBUG_HTTP, "Got 0 status, changing to 1 for safety.\n" );
	req->status = 1;
    }
	
    return ret;
}

/* Simple HTTP put. 
 * local is the local filename. Remote is the destination URI (URI?)
 * Make it proper.
 * Returns:
 *   PROTO_FILE if no local file
 *   PROTO_ERROR if something general goes wrong
 *   PROTO_OK if it all works fine
 */
int http_put( const char *local, const char *remote, const bool ascii ) {
    http_req_t req;
    int ret;
    
    http_request_init( &req, "PUT", remote );

    req.body_file = fopen( local, "r" );
    if( req.body_file == NULL ) {
	strcpy( http_error, "Could not open file." );
	return PROTO_FILE;
    }
    req.body = http_body_file;

    ret = http_request( &req );
    fclose( req.body_file );

    if( ret!=PROTO_OK ) return PROTO_ERROR;

    if( req.status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

int http_get_fd;
bool http_get_working;
size_t http_get_total, http_get_progress;

void http_get_callback( const char *buffer, const size_t len ) {
    if( !http_get_working ) return;
    if( send_data( http_get_fd, buffer, len ) < 0 ) {
	http_get_working = false;
    } else {
	http_get_progress += len;
    }
    fe_transfer_progress( http_get_progress, http_get_total ); 
}

int http_get( const char *local, const char *remote, const int remotesize,
	      const bool ascii ) {
    http_req_t req;
    int ret;

#ifdef __EMX__
    /* We have to set O_BINARY, thus need open(). Otherwise it should be
       equivalent to creat(). */
    http_get_fd = open( local, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0644 );
#else
    http_get_fd = creat( local, 0644 );
#endif
    if( http_get_fd < 0 ) {
	snprintf( http_error, BUFSIZ, "Could not open local file: %s", 
		  strerror( errno ) );
	return PROTO_ERROR;
    }

    http_request_init( &req, "GET", remote );
    req.body_callback = http_get_callback;
    http_get_working = true;
    req.resp_body_size = &http_get_total;
    http_get_progress = 0;

    ret = http_request( &req );
    
    if( close( http_get_fd ) < 0 ) {
	snprintf( http_error, BUFSIZ, "Error closing local file: %s",
		  strerror( errno ) );
	ret = PROTO_ERROR;
    }

    if( ret != PROTO_OK )
	return PROTO_ERROR;

    if( req.status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

/* Perform the file operations */
int dav_move( const char *from, const char *to ) {
    http_req_t req;

    http_request_init( &req, "MOVE", from );

    strcat( req.headers, "Destination: http://" );
    http_strcat_remoteserver( req.headers );
    strcat( req.headers, to );
    /* For non-WebDAV servers */
    strcat( req.headers, EOL "New-URI: " );
    strcat( req.headers, to );
    strcat( req.headers, EOL );

    if( http_request( &req ) != PROTO_OK )
	return PROTO_ERROR;	
    
    if( req.status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

/* Deletes the specified resource on the server */
int http_delete( const char *filename ) {
    http_req_t req;
    
    http_request_init( &req, "DELETE", filename );

    if( http_request( &req ) != PROTO_OK )
	return PROTO_ERROR;

    if( req.status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

/* Deletes the specified resource on the server */
int dav_rmdir( const char *filename ) {
    http_req_t req;
    char *dirname;
    
    if( strlen( filename ) < 1 ) {
	return PROTO_ERROR;
    }

    if( *(filename+strlen(filename)) != '/' ) {
	dirname = malloc( strlen( filename ) + 2 );
	strcpy( dirname, filename );
	strcat( dirname, "/" );
    } else {
	dirname = (char *)filename;
    }

    http_request_init( &req, "DELETE", dirname );
    
    if( http_request( &req ) != PROTO_OK )
	return PROTO_ERROR;

    if( req.status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

int http_mkdir_with_put( const char *realdir ) {
    char filename[BUFSIZ];
    http_req_t req;

    strcpy( filename, realdir );
    strcat( filename, "SitecopyTempDirCreator.txt" );

    http_request_init( &req, "PUT", filename );
    /* Can't send Content-Type, would break digest auth entity
     * digest code
    strcat( req.headers, "Content-type: text/plain" EOL );
     */
    
    if( http_request( &req ) != PROTO_OK )
	return PROTO_ERROR;

    if( req.status/100 == 2 ) {
	http_request_init( &req, "DELETE", filename );

        if( (http_request( &req ) != PROTO_OK ) || ( req.status/100 != 2 ) )
	    /* NEED an FE warning function here */
            DEBUG( DEBUG_HTTP,
		   "Can\'t delete the temporary file http://%s:%d%s",
		   http_remotehost, http_remoteport, filename );
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }
}

int dav_mkcol( const char *dirname ) {
    http_req_t req;
    char *realdir;

    if( strlen( dirname ) < 1 ) {
	strcpy( http_error, "Invalid directory name." );
	return PROTO_ERROR;
    }
    if( *(dirname+strlen(dirname)-1) == '/' ) {
	realdir = (char *) dirname;
    } else {
	/* +2 since one for \0, one for / */
	realdir = malloc( strlen(dirname) + 2 );
	strcpy( realdir, dirname );
	strcat( realdir, "/" );
    }
    
    if( http_mkdir_works == 1 ) {
	/* Use MKDIR, since we know it works */

	http_request_init( &req, "MKDIR", realdir );
	if( http_request( &req ) != PROTO_OK )
	    return PROTO_ERROR;
	if( req.status/100 ) {
	    return PROTO_OK;
	} else {
	    return PROTO_ERROR;
	}

    } else {
	/* Try MKCOL first */
	http_request_init( &req, "MKCOL", realdir );
	
	if( http_request( &req ) != PROTO_OK )
	    return PROTO_ERROR;
	
	if( req.status/100 == 2 ) {
	    /* MKCOL works -> MKDIR doesn't */
	    http_mkdir_works = -1;
	    return PROTO_OK;
	} else if( (http_mkdir_works > -1) && !http_webdav_server ) { 
	    /* MKCOL failed, we're not on a DAV server, lets try MKDIR */
	    http_request_init( &req, "MKDIR", realdir );
	    if( http_request( &req ) != PROTO_OK )
		return PROTO_ERROR;
	    if( req.status/100 == 2 ) {
		/* MKDIR does work */
		http_mkdir_works = 1;
		return PROTO_OK;
	    } else { 
		/* MKDIR doesn't work */
		if( req.status == 501 ) {
		    /* MKDIR definitely isn't implemented */
		    http_mkdir_works = -1;
		}
		/* Try a dummy PUT/DELETE */
		return http_mkdir_with_put( realdir );
	    }
	} else {
	    return PROTO_ERROR;
	}
    }
}

int http_open( void ) {
    DEBUG( DEBUG_SOCKET, "Connecting to %s:%d...\n", http_remotehost, http_remoteport );
    http_sock = socket_connect( http_remoteaddr, http_remoteport );
    if( http_sock < 0 ) {
	DEBUG( DEBUG_SOCKET, "Could not connect: %s\n", strerror( errno ) );
	return PROTO_CONNECT;
    }
    DEBUG( DEBUG_SOCKET, "Connected.\n" );
    http_connected = true;
    return PROTO_OK;
}

int http_close( void ) {
    DEBUG( DEBUG_SOCKET, "Closing socket.\n" );
    socket_close( http_sock );
    http_connected = false;
    DEBUG( DEBUG_SOCKET, "Socket closed.\n" );
    return PROTO_OK;
}

int dav_mkref( const char *resref, const char *target ) {
    http_req_t req;

    http_request_init( &req, "MKREF", resref );
    strcat( req.headers, "Ref-Target: <" );
    strcat( req.headers, target );
    strcat( req.headers, ">" EOL );

    if( http_request( &req ) != PROTO_OK )
	return PROTO_ERROR;

    if( req.status/100 == 2 ) {
	return PROTO_OK;
    }
    
    return PROTO_ERROR;
    
}

int dav_chref( const char *resref, const char *target ) {
    int ret;

    /* Delete it */
    ret = dav_rmref( resref );
    if( ret != PROTO_OK ) 
	return ret;
    /* Then create it again */
    ret = dav_mkref( resref, target );

    return ret;
}

int dav_rmref( const char *resref ) {
    http_req_t req;
    
    http_request_init( &req, "DELETE", resref );
    strcat( req.headers, "No-Passthrough: 1" EOL );
    
    if( http_request( &req ) != PROTO_OK )
	return PROTO_ERROR;

    if( req.status/100 == 2 ) {
	return PROTO_OK;
    }
    
    return PROTO_ERROR;
}

#ifdef HAVE_LIBEXPAT

/* TODO: First time hack. Bit carp having dav_xml_state global.
 * Do it better, pass something round using expat's userdata. Go
 * read the XML specs again and use the correct terminology for 
 * everything.
 * TODO: free() everything when we're done, atm, everything leaks.
 */

typedef enum {
    dav_xml_multistatus,
    dav_xml_response,
    dav_xml_href,
    dav_xml_propstat,
    dav_xml_prop,
    dav_xml_getlastmodified,
    dav_xml_getcontentlength,
    dav_xml_resourcetype,
    dav_xml_status,
    dav_xml_responsedescription,
    dav_xml_collection,
    dav_xml_unknown,
    dav_xml_root
} dav_xml_tag_t;

typedef struct  {
    const char *name;
    dav_xml_tag_t tag;
} dav_xml_tag_map_t;

/* TODO: unnecessary, just have an array of tag names where the
 * order corresponds to the tag number */
dav_xml_tag_map_t dav_xml_tagmap[] = {
    { "DAV:multistatus", dav_xml_multistatus },
    { "DAV:response", dav_xml_response },
    { "DAV:responsedescription", dav_xml_responsedescription },
    { "DAV:href", dav_xml_href },
    { "DAV:propstat", dav_xml_propstat },
    { "DAV:prop", dav_xml_prop },
    { "DAV:status", dav_xml_status },
    { "DAV:getlastmodified", dav_xml_getlastmodified },
    { "DAV:getcontentlength", dav_xml_getcontentlength },
    { "DAV:resourcetype", dav_xml_resourcetype },
    { "DAV:collection", dav_xml_collection }
};

const int dav_xml_tagmap_count = sizeof(dav_xml_tagmap) / sizeof(dav_xml_tag_map_t);

typedef struct dav_xml_ns_s dav_xml_ns;

/* Linked list of namespace scopes */
struct dav_xml_ns_s {
    char *name;
    char *value;
    dav_xml_ns *next;
};

typedef struct {
    char *href;
    time_t modtime;
    size_t size;
    bool isdir;
} dav_xml_fileinfo;

typedef struct dav_xml_state_s dav_xml_state_t;

struct dav_xml_state_s {
    dav_xml_tag_t tag;
    char *tag_name;
    char *default_ns;
    dav_xml_ns *nspaces; /* linked list */
    dav_xml_state_t *parent;
    char cdata[BUFSIZ]; /* FIXME: this will break */
    bool want_cdata; /* whether to collect the cdata for this element */
    dav_xml_fileinfo *file; /* for storing file info */
};

dav_xml_state_t *dav_xml_state;
struct proto_file_t *dav_fetch_list;
const char *dav_fetch_root;

XML_Parser dav_xml_parser;
bool dav_xml_parsing;

/* Parses the tag attributes, and handles XML namespaces. 
 * With a little bit of luck.
 * Return true on success. */
bool dav_xml_parsetag( const char *tag, const char **atts ) {
    dav_xml_ns *ns;
    int attn;
    char *pnt, *tmp;
    const char *tag_prefix, *tag_suffix;
    dav_xml_state_t *xmlt;
    DEBUG( DEBUG_XML, "Parsing tag of name: [%s]\n", tag );
    /* Parse the atts for namespace declarations */
    for( attn = 0; atts[attn]!=NULL; attn+=2 ) {
	DEBUG( DEBUG_XML, "Got attribute: [%s] = [%s]\n", atts[attn], atts[attn+1] );
	if( strcasecmp( atts[attn], "xmlns" ) == 0 ) {
	    /* New default namespace */
	    dav_xml_state->default_ns = strdup( atts[attn+1] );
	    DEBUG( DEBUG_XML, "New default namespace: %s\n", 
		   dav_xml_state->default_ns );
	} else if( strncasecmp( atts[attn], "xmlns:", 6 ) == 0 ) {
	    /* New namespace scope */
	    ns = malloc( sizeof( dav_xml_ns ) );
	    ns->next = dav_xml_state->nspaces;
	    dav_xml_state->nspaces = ns;
	    ns->name = strdup( atts[attn]+6 );
	    ns->value = strdup( atts[attn+1] );
	    DEBUG( DEBUG_XML, "New namespace scope: %s -> %s\n",
		   ns->name, ns->value );
	}
    }
    /* Now check the tag name for a namespace scope */
    pnt = strchr( tag, ':' );
    tag_prefix = NULL;
    tag_suffix = NULL;
    if( pnt == NULL ) {
	/* No scope - have we got a default? */
	DEBUG( DEBUG_XML, "No scope found, searching for default.\n" );
	for( xmlt = dav_xml_state; xmlt!=NULL; xmlt=xmlt->parent ) {
	    if( xmlt->default_ns != NULL ) {
		tag_prefix = xmlt->default_ns;
		break;
	    }
	}
	if( tag_prefix != NULL ) {
	    DEBUG( DEBUG_XML, "Found default namespace [%s]\n", tag_prefix );
	} else {
	    DEBUG( DEBUG_XML, "No default namespace, using empty.\n" );
	    tag_prefix = "";
	}
	tag_suffix = tag;
    } else {
	/* hacky hacky hacky */
	tmp = strdup( tag );
	strtok( tmp, ":" );
	DEBUG( DEBUG_XML, "Got namespace scope [%s]. Trying to resolve...",
	       tmp );
	/* Have a scope - resolve it */
	for( xmlt = dav_xml_state; tag_prefix==NULL && xmlt!=NULL; xmlt=xmlt->parent ) {
/*	    DEBUG( DEBUG_XML, "\nNew element. Comparing:" ); */
	    for( ns = xmlt->nspaces; ns!=NULL; ns=ns->next ) {
/*		DEBUG( DEBUG_XML, " %s", ns->name ); */
		if( strcasecmp( ns->name, tmp ) == 0 ) {
		    /* Scope matched! Hoorah */
/*		    DEBUG( DEBUG_XML, " - MATCHED.\n" ); */
		    tag_prefix = ns->value;
		    /* end the search */
		    break;
		}
	    }
	}
	if( tag_prefix != NULL ) {
	    DEBUG( DEBUG_XML, "Resolved scope [%s] to [%s]\n", 
		   tmp, tag_prefix );
	    /* Chuck the prefix, we need the element name now */
	    tmp = strtok( NULL, ":" );
	    tag_suffix = strdup( tmp );
	} else {
	    DEBUG( DEBUG_XML, "Smeg. We lost it somewhere.\n" );
	    return false;
	}
    }
    /* here, we have tag_suffix and tag_prefix */
    DEBUG( DEBUG_XML, "prefix: [%s], suffix: [%s]\n", tag_prefix, tag_suffix );
    pnt = dav_xml_state->tag_name = malloc( BUFSIZ ); /* FIXME: lazy me */
    strcpy( pnt, tag_prefix );
    strcat( pnt, tag_suffix );
    DEBUG( DEBUG_XML, "You gave me: %s, and I gave you this: %s\n", tag, pnt );
    return true;
}


void dav_xml_startelm( void *userdata, const char *tag, const char **atts ) {
    /* Allocate a new frame */
    dav_xml_state_t *s;
    int n;

    /* Set the new state */
    s = malloc( sizeof(dav_xml_state_t) );
    memset( s, 0, sizeof(dav_xml_state_t) );
    s->parent = dav_xml_state;
    dav_xml_state = s;
    if( dav_xml_parsetag( tag, atts ) == false ) {
	/* nooo, it bombed */
	dav_xml_parsing = false;
	return;
    }
    /* Map the name to a tag */
    DEBUG( DEBUG_XML, "Mapping tag name... " );
    s->tag = dav_xml_unknown;
    for( n = 0; n < dav_xml_tagmap_count; n++ ) {
	if( strcasecmp( dav_xml_tagmap[n].name, s->tag_name ) == 0 ) {
	    s->tag = dav_xml_tagmap[n].tag;
	    break;
	}
    }
    if( s->tag == dav_xml_unknown ) {
	DEBUG( DEBUG_XML, "unknown tag, ignoring.\n" );
	return;
    }
    DEBUG( DEBUG_XML, "mapped.\n" );
    /* Be snooty. They ain't giving us nothing we don't understand. */
    switch( s->parent->tag ) {
    case dav_xml_root:
	switch( s->tag ) {
	case dav_xml_multistatus:
	case dav_xml_response:
	    break;
	default:
	    dav_xml_parsing = false;
	    break;
	}
	break;
    case dav_xml_multistatus:
	switch( s->tag ) {
	case dav_xml_response:
	case dav_xml_responsedescription:
	    break;
	default:
	    dav_xml_parsing = false;
	    break;
	}
	break;
    case dav_xml_response:
	switch( s->tag ) {
	case dav_xml_href:
	case dav_xml_propstat:
	case dav_xml_responsedescription:
	case dav_xml_status:
	    break;
	default:
	    dav_xml_parsing = false;
	}
	break;
    default:
	break;
    }
    if( dav_xml_parsing == false ) {
	DEBUG( DEBUG_XML, "Parse error.\n" );
	return;
    }
    /* Now actually do something about it */
    switch( s->tag ) {
    case dav_xml_response:
	/* new file information */
	DEBUG( DEBUG_XML, "New file context.\n" );
	s->file = malloc( sizeof(dav_xml_fileinfo) );
	memset( s->file, 0, sizeof(dav_xml_fileinfo) );
	break;

    case dav_xml_href:
    case dav_xml_getcontentlength:
    case dav_xml_resourcetype:
    case dav_xml_getlastmodified:
	/* For these elements, collect the CDATA */
	DEBUG( DEBUG_XML, "Collecting CDATA...\n" );
	s->want_cdata = true;
	s->file = s->parent->file;
	break;
    case dav_xml_prop:
    case dav_xml_propstat:
    case dav_xml_collection:
	/* let the child tags have access to the same fileinfo */
	s->file = s->parent->file;
	break;
    default:
	/* donothing */
	break;
    }
}

/* Returns the depth of the file in the directory heirarchy,
 * i.e. 1 for /foo.html, 2 for /bar/norm.html, 4 for /a/b/c/d.html
 */
int dav_fetch_getdepth( const char *href ) {
    const char *pnt;
    int count = 0;
    for( pnt=href; *pnt != '\0'; pnt++ ) /* oneliner */
	if( *pnt == '/' ) count++;
    return count;
}

char *dav_fetch_parse_href( const char *href ) {
    char *tmp, *ret;
    DEBUG( DEBUG_HTTP, "Parsing href [%s]\n", href );
    if( strncmp( href, "http://", 7 ) == 0 ) {
	/* Absolute URI.
	 * Look for the path bit */
	DEBUG( DEBUG_HTTP, "Got absolute URI.\n" );
	tmp = strchr( href+7, '/' );
	if( tmp == NULL ) {
	    DEBUG( DEBUG_HTTP, "No path segment found.\n" );
	    return NULL;
	}
	DEBUG( DEBUG_HTTP, "Got path segment: %s\n", tmp );
	if( strncmp( tmp, dav_fetch_root, strlen( dav_fetch_root ) ) != 0 ) {
	    DEBUG( DEBUG_HTTP, "Doesn't match our root.\n" );
	    return NULL;
	}
    } else {
	tmp = (char *) href;
    }
    ret = strdup( tmp + strlen( dav_fetch_root ) );
    DEBUG( DEBUG_HTTP, "parse_href: Returning [%s]\n", ret );
    return ret;
}

/* TODO: Redo fetch mode properly, this is arse */
void dav_fetch_gotresource( dav_xml_state_t *s) {
    struct proto_file_t *file, *current, *previous;
    char *href_rel;

    DEBUG( DEBUG_HTTP, "Got resource:\n"
	   "href = [%s] isdir = %s size = %d mtime = %s\n", s->file->href,
	   s->file->isdir?"true":"false", s->file->size, rfc1123_date( s->file->modtime ) );

    /* Work out the href relative to the site root */
    href_rel = dav_fetch_parse_href( s->file->href );
    if( href_rel == NULL ) {
	DEBUG( DEBUG_HTTP, "Could not parse href, ignoring resoure.\n" );
	return;
    }
    DEBUG( DEBUG_HTTP, "Relative href is: %s\n", href_rel );
    if( strlen( href_rel ) == 0 ) {
	DEBUG( DEBUG_HTTP, "Resource is root collection, ignoring.\n" );
	return;
    }
   
    if( s->file->isdir ) {
	href_rel[strlen(href_rel)-1] = '\0';
	DEBUG( DEBUG_HTTP, "Relative href is really: %s\n", href_rel );
    }

    file = malloc( sizeof( struct proto_file_t ) );

    file->isdir = s->file->isdir;
    file->directory = strdup( dirname( href_rel ) );
    file->filename = basename( href_rel );
    file->modtime = s->file->modtime;
    file->size = s->file->size;
    file->depth = dav_fetch_getdepth( href_rel );
    
    DEBUG( DEBUG_HTTP, "Resource: dir: [%s], filename [%s]\n",
	   file->directory, file->filename );

    /* Insert it into the list, keeping the list sorted by depth */

    for( current=dav_fetch_list, previous=NULL; current!=NULL; 
	 previous=current, current=current->next )
	/* one-liner */
	if( current->depth > file->depth )
	    break;

    file->next = current;

    if( previous == NULL ) {
	dav_fetch_list = file;
    } else {
	previous->next = file;
    }

}

time_t http_dateparse( const char *date ) {
    time_t tmp;
    tmp = rfc1123_parse( date );
    if( tmp == -1 )
	tmp = rfc1036_parse( date );
    return tmp;
}

#define XML_WHITESPACE "\r\n\t "
/* Cleans the CDATA of leading and trailing whitespace */
void dav_xml_wsclean( dav_xml_state_t *s ) {
    int count;
    char *tmp;
    DEBUG( DEBUG_XML, "Cleaning CDATA.\n" );
    count = strspn( s->cdata, XML_WHITESPACE );
    if( count > 0 ) { 
	if( count == strlen( s->cdata ) ) {
	    /* The cdata is ALL whitespace */
	    DEBUG( DEBUG_XML, "cdata is all whitespace." );
	    s->cdata[0] = '\0';
	} else {
	    DEBUG( DEBUG_XML, "Cleaning %d bytes from start.\n", count );
	    tmp = malloc( strlen(s->cdata) - count + 1 );
	    memcpy( tmp, s->cdata+count, strlen(s->cdata)-count+1 );
	    DEBUG( DEBUG_XML, "Before:\n%s\nAfter:\n%s\n", s->cdata, tmp );
	    strcpy( s->cdata, tmp );
	    free( tmp );
	}
    } else {
	DEBUG( DEBUG_XML, "No whitespace to clean.\n" );
    }
}

/* End-of-element handler */
void dav_xml_endelm( void *userdata, const char *tag ) {
    dav_xml_state_t *s = dav_xml_state;
    if( dav_xml_parsing == false )
	return;
    DEBUG( DEBUG_XML, "End of the element %s.\n", tag );
    if( s->want_cdata ) {
	dav_xml_wsclean( s );
    }
    switch( s->tag ) {
    case dav_xml_href:
	s->file->href = s->cdata;
	break;
    case dav_xml_getlastmodified:
	s->file->modtime = http_dateparse( s->cdata );
	break;
    case dav_xml_getcontentlength:
	s->file->size = atoi( s->cdata );
	break;
    case dav_xml_collection:
	s->file->isdir = true;
	break;
    case dav_xml_response:
	dav_fetch_gotresource( s );
	break;
    default:
	break;
    }
    dav_xml_state = dav_xml_state->parent;
}

/* CDATA handler. */
void dav_xml_cdata( void *userdata, const char *cdata, int len ) {
    char *tmp;
    if( dav_xml_state->want_cdata ) {
	tmp = malloc( len+1 );
	strncpy( tmp, cdata, len );
	tmp[len] = '\0';
	DEBUG( DEBUG_XML, "Got cdata: [%s]\n", tmp );
	strcat( dav_xml_state->cdata, tmp );
    }
}

void dav_xml_parsebody( const char *buffer, const size_t len ) {
    int ret;
    /* duck out if it's broken */
    if( !dav_xml_parsing ) {
	DEBUG( DEBUG_XML, "Not parsing %d bytes!\n", len );
	return;
    }
    if( len == 0 ) {
	DEBUG( DEBUG_XML, "Got 0-length buffer, end of response.\n" );
	ret = XML_Parse( dav_xml_parser, "", 0, -1 );
    } else {
	DEBUG( DEBUG_XML, "Got %d length buffer.\n", len );
	ret = XML_Parse( dav_xml_parser, buffer, len, 0 );
    }
    DEBUG( DEBUG_XML, "XML_Parse returned %d\n", ret );
    if( ret == 0 ) {
	dav_xml_parsing = false;
    }
}

/* Passes the file list back up */
void dav_fetch_passup( struct proto_file_t *list, fetch_func_t callback ) {
    struct proto_file_t *file;
    for( file = list; file!=NULL; file=file->next ) {
	(*callback)( file->filename, file->directory, file->isdir, 
		     file->size, file->modtime );
    }
}
   

/* TODO: Is this utf-8? */
int dav_fetch( const char *dirname, fetch_func_t callback ) {
    http_req_t req;
    int ret;
    const char *propfind_body =
	"<?xml version=\"1.0\"?>" EOL /* should we use encoding=? */
	"<propfind xmlns=\"DAV:\">" EOL
	"  <prop>" EOL
	"    <getcontentlength/>" EOL
	"    <getlastmodified/>" EOL
	"    <resourcetype/>" EOL
	"  </prop>" EOL
	"</propfind>" EOL;
    const char *myheaders =
	"Content-Type: text/xml" EOL /* should we use charset=? */
	"Depth: infinity" EOL;
       
    dav_xml_parser = XML_ParserCreate( NULL );
    XML_SetElementHandler( dav_xml_parser, dav_xml_startelm, dav_xml_endelm );
    XML_SetCharacterDataHandler( dav_xml_parser, dav_xml_cdata );
   
    /* Initialize */
    dav_xml_state = malloc( sizeof( dav_xml_state_t ) );
    memset( dav_xml_state, 0, sizeof( dav_xml_state_t ) );
    dav_xml_state->tag = dav_xml_root;
    dav_fetch_list = NULL;
    dav_fetch_root = dirname;

    http_request_init( &req, "PROPFIND", dirname );
    req.body_callback = dav_xml_parsebody;
    req.body = http_body_buffer;
    req.body_buffer = propfind_body;
    /* Add in the content type header */
    strcat( req.headers, myheaders );
    dav_xml_parsing = true;
    
    ret = http_request( &req );

    XML_ParserFree( dav_xml_parser );

    dav_fetch_passup( dav_fetch_list, callback );

    if( (ret == PROTO_OK) && dav_xml_parsing ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }
}

#endif /* HAVE_LIBEXPAT */

int http_head( const char *directory ) {
    http_req_t req;
    
    http_request_init( &req, "HEAD", directory );
    if( http_request( &req ) != PROTO_OK )
	return PROTO_ERROR;

    if( (req.status / 100) == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

void http_options_parsehdr( const char *name, const char *value ) {
    if( strcasecmp( name, "DAV" ) == 0 ) {
	/* TODO: Parse it properly */
	DEBUG( DEBUG_HTTP, "Got DAV header with value: %s\n", value );
	if( strncmp( value, "1", 1 ) == 0 ) {
	    DEBUG( DEBUG_HTTP, "Class 1 compliant server.\n" );
	    http_webdav_server = true;
	}
    }
}

/* Performs an OPTIONS request.
 * Sets http_webdav_server appropriately.
 */
int http_options( const char *directory ) {
    http_req_t req;
    
    http_webdav_server = false;

    http_request_init( &req, "OPTIONS", directory );
    req.hdrs_callback = http_options_parsehdr;

    if( http_request( &req ) != PROTO_OK )
	return PROTO_ERROR;
    
    if( req.status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}
